!--------------------------------------------------------------------
program checkoutput
!--------------------------------------------------------------------
! test if output source counts and rates deviate by more than 5 %
! from reference data set
! test if output source position deviate by more than 1 pixel from 
! reference data set

use dal
use errorhandling

implicit none

integer(kind=int32) :: i, j, k,itest, nrow1, nrow2
type(DataSetT)  :: set1, set2
type(TableT)    :: tab1, tab2
Type(ColumnT)   :: scol1, scol2, rcol1, rcol2, xcol1, xcol2, ycol1, ycol2
real(kind=single), dimension(:), pointer :: scts1, scts2, rate1, rate2, &
                   xima1, xima2, yima1, yima2
real(kind=single) ::  maxcts, upcts
integer :: ok,ntest,imax
character(len=20), dimension(3) :: &
          list = (/"eboxlist1.fits","eboxlist2.fits","eboxlist3.fits"/), &
          ref  = (/"eboxlist1.ref","eboxlist2.ref","eboxlist3.ref"/)

integer, dimension(10) :: isort1,isort2      

do itest = 1, 3

  set1 = dataSet(list(itest), read)
  tab1 = table(set1,"SRCLIST")
  nrow1 = numberOfRows(tab1)
  scol1 = column(tab1,"SCTS", read)
  scts1 => real32Data(scol1)
  rcol1 = column(tab1,"RATE", read)
  rate1 => real32Data(rcol1)
  xcol1 = column(tab1,"X_IMA", read)
  xima1 => real32Data(xcol1)
  ycol1 = column(tab1,"Y_IMA", read)
  yima1 => real32Data(ycol1)

  set2 = dataSet(ref(itest), read)
  tab2 = table(set2,"SRCLIST")
  nrow2 = numberOfRows(tab2)
  scol2 = column(tab2,"SCTS", read)
  scts2 => real32Data(scol2)
  rcol2 = column(tab2,"RATE", read)
  rate2 => real32Data(rcol2)
  xcol2 = column(tab2,"X_IMA", read)
  xima2 => real32Data(xcol2)
  ycol2 = column(tab2,"Y_IMA", read)
  yima2 => real32Data(ycol2)


  ntest=min(10,nrow1-1,nrow2-1)
  maxcts=0.0  
  upcts=1.E10  
  do i=1,ntest
    do j=0,nrow1-1
      if (scts1(j) > maxcts .and. scts1(j) < upcts) then
        imax=j
        maxcts=scts1(j) 
      end if     
    end do
    isort1(i)=imax
    upcts=maxcts 
  end do   

  maxcts=0.0  
  upcts=1.E10  
  do i=1,ntest
    do j=0,nrow2-1
      if (scts2(j) > maxcts .and. scts2(j) < upcts) then
        imax=j
        maxcts=scts2(j) 
      end if     
    end do
    isort2(i)=imax
    upcts=maxcts 
  end do     


  ok = 0
  do i = 1,ntest
    j=isort1(i)
    k=isort2(i) 
    if (scts1(j) < 0.90*scts2(k) .or. &
        scts1(j) > 1.10*scts2(k)) then
      ok = ok+1
      print *,"SCTS:",itest,i,scts1(j),scts2(k)
    end if
    if (rate1(j) < 0.90*rate2(k) .or. &
        rate1(j) > 1.10*rate2(k)) then
      ok = ok+1
      print *,"RATE:",itest,i,rate1(j),rate2(k)
    end if 
    if (xima1(j) < xima2(k)-1.0 .or. &
        xima1(j) > xima2(k)+1.0) then
      ok = ok+1
      print *,"XIMA:",itest,i,xima1(j),xima2(k)
    end if  
    if (yima1(j) < yima2(k)-1.0 .or. &
        yima1(j) > yima2(k)+1.0) then
      ok = ok+1
      print *,"YIMA:",itest,i,yima1(j),yima2(k)
    end if 
  end do

  call release(scol1)
  call release(scol2)
  call release(rcol1)
  call release(rcol2)
  call release(xcol1)
  call release(xcol2)
  call release(ycol1)
  call release(ycol2)
  call release(tab1)
  call release(tab2)
  call release(set1)
  call release(set2)

end do

if (ok > 5) call error("test failed","source list contains wrong values")

end program checkoutput

